/*
 * roi.hpp
 *
 *     Project: jafar
 *  Created on: Jun 1, 2010
 *      Author: jsola
 */

#ifndef ROI_HPP_
#define ROI_HPP_

#include <cmath>
#include <boost/shared_ptr.hpp>

#include "jmath/jblas.hpp"

#include "image/Image.hpp"
#include "image/imageException.hpp"

#define USE_TEMPLATE_INLINE 0

namespace jafar {
	namespace image {
		using namespace jblas;

//		class RoiAbstract;
//		typedef boost::shared_ptr<RoiAbstract> roi_ptr_t;
/*
Well well... so many solutions...
Do we want a very general ROI, or is convex ROI enough ? with a convex roi some operations can fail (remove, add...)
For a general ROI, do we use a mask (meaning we have to test every pixel), or a slice model ?
For a slice model, do we use a vector of list, just a list, both ?
Do we use std::list, or a custom list which stores its data in an array to optimize memory management ?

Let's start with simple convex roi, while  thinking to a general roi




*/

//#define NEW_ROI

		/**
		This class represents a convex 2D Region Of Interest.
		*/
		
		struct ConvexSlice { int x, w; };
		std::ostream& operator<<(std::ostream &os, ConvexSlice const &c);

		class ConvexRoi
		{
			public:
				typedef std::vector<ConvexSlice> ConvexRoiImpl;
			private:
				cv::Rect rect; // bounding box
				ConvexRoiImpl convex; // convex area, with origin at rect.x,rect.y
				bool hasConvex;
				int count_;
			private:
				void initConvex()
				{
					convex.resize(rect.height);
					for(ConvexRoiImpl::iterator it = convex.begin(); it != convex.end(); ++it)
					{
						it->x = 0;
						it->w = rect.width;
					}
					hasConvex = true;
				}
			public:
				// constructors
				ConvexRoi(): rect(cv::Rect(0,0,0,0)), hasConvex(false), count_(0) {}
				ConvexRoi(cv::Rect &rect): rect(rect), convex(), hasConvex(false), count_(rect.width*rect.height) {}
				ConvexRoi(int x, int y, int width, int height): rect(x,y,width,height), convex(), hasConvex(false), count_(width*height) {}
				ConvexRoi(jblas::vec2 x, jblas::sym_mat22 P, double scale);
				
				void init(cv::Rect _rect) { rect = _rect; convex.resize(0); hasConvex = false; count_ = rect.width*rect.height; }
				
				// accessors
				int y(     ) const { return rect.y; }                                  ///< roi and bounding box top
				int h(     ) const { return rect.height; }                             ///< roi and bounding box height
				int x(     ) const { return rect.x; }                                  ///< bounding box left
				int x(int i) const { return rect.x + (hasConvex ? convex[i-rect.y].x : 0); }  ///< roi left
				int w(     ) const { return rect.width; }                              ///< bounding box width
				int w(int i) const { return hasConvex ? convex[i-rect.y].w : rect.width; }    ///< roi width
				int size(  ) const { return w()*h(); }                                 ///< roi size
				int count( ) const { return count_; }                                  ///< roi elements count
				bool isIn(jblas::vec2 p) const
				{
					int py = std::floor(p(1));
					int px = std::floor(p(0));
					return (py >= y() && py < y()+h() && px >= x(py) && px < x(py)+w(py));
				}
				
				/// union
				ConvexRoi  operator+ (const ConvexRoi &roi) const 
				{
					ConvexRoi res(*this);
					// TODO
					JFR_ERROR(ImageException, ImageException::NOT_IMPLEMENTED, "this operator hasn't been implemented yet");
					return res;
				}
				
				/// union
				ConvexRoi& operator+=(const ConvexRoi &roi)
				{
					// TODO
					JFR_ERROR(ImageException, ImageException::NOT_IMPLEMENTED, "this operator hasn't been implemented yet");
					return *this;
				}
				
				/// difference (A and not B)
				ConvexRoi  operator- (const ConvexRoi &roi) const
				{
					ConvexRoi res(*this);
					// TODO
					JFR_ERROR(ImageException, ImageException::NOT_IMPLEMENTED, "this operator hasn't been implemented yet");
					return res;
				}
				
				/// difference (A and not B)
				ConvexRoi& operator-=(const ConvexRoi &roi)
				{
					// TODO
					JFR_ERROR(ImageException, ImageException::NOT_IMPLEMENTED, "this operator hasn't been implemented yet");
					return *this;
				}
				
				/// intersection
				ConvexRoi  operator* (const ConvexRoi &roi) const
				{
					// first deal with the rects
					cv::Rect newRect;
					newRect.x = std::max(rect.x, roi.rect.x);
					int x2 = std::min(rect.x+rect.width-1, roi.rect.x+roi.rect.width-1);
					newRect.width = std::max(0, x2-newRect.x+1);
					newRect.y = std::max(rect.y, roi.rect.y);
					int y2 = std::min(rect.y+rect.height-1, roi.rect.y+roi.rect.height-1);
					newRect.height = std::max(0, y2-newRect.y+1);
					// now with convex
					if (!hasConvex && !roi.hasConvex)
					{
						return ConvexRoi(newRect);
					} else
					if (hasConvex && roi.hasConvex)
					{
						// TODO
						JFR_ERROR(ImageException, ImageException::NOT_IMPLEMENTED, "this operator hasn't been implemented yet");
					} else
					{
						ConvexRoi newRoi(newRect);
						newRoi.initConvex();
						newRoi.count_ = 0;
						
						const cv::Rect &oldRect = (hasConvex ? roi.rect : rect);
						const ConvexRoiImpl &oldConvex = (hasConvex ? convex : roi.convex);
						const cv::Rect &oldConvexRect = (hasConvex ? rect : roi.rect);
						
						int shift = std::max(0, oldRect.y - oldConvexRect.y);
						int lastY = -1000000;
						int minX1 = 1000000, maxX2 = -1000000;
//JFR_DEBUG("newRect begin " << newRect);
						for(int y = 0; y < newRect.height; ++y)
						{
							ConvexSlice &newSlice = newRoi.convex[y];
							ConvexSlice const &oldSlice = oldConvex[y+shift];
							newSlice.x = std::max(newRect.x, oldConvexRect.x+oldSlice.x)-newRect.x;
							int newX2 = std::min(newRect.x+newRect.width-1, oldConvexRect.x+oldSlice.x+oldSlice.w-1);
							newSlice.w = newX2-(newRect.x+newSlice.x)+1;
//JFR_DEBUG("slice y=" << y << " " << newSlice);
							if (newSlice.x >= newRect.width || newSlice.w <= 0)
							{
								if (lastY == -1000000)
								{
									newRect.y++; newRect.height--;
									shift++;
									y--;
									continue;
								} else
								{
									newSlice.x = 0;
									newSlice.w = 0;
								}
//JFR_DEBUG("error branch: newRect " << newRect << ", shift " << shift);
							} else
							{
								lastY = y;
								newRoi.count_ += newSlice.w;
								if (newSlice.x < minX1) minX1 = newSlice.x;
								if (newX2 > maxX2) maxX2 = newX2;
//JFR_DEBUG("no error branch: lastY " << lastY << ", minX1 " << minX1 << ", maxX2 " << maxX2);
							}
						}
						if (lastY == -1000000)
						{
							newRoi.rect.height = 0;
							newRoi.convex.resize(0);
						} else
						{
							newRoi.rect.height = lastY+1;
							newRoi.rect.width = maxX2-minX1-newRect.x+1;
							newRoi.rect.x += minX1;
							newRoi.convex.resize(newRoi.rect.height);
							if (minX1 != 0) for(int y = 0; y < newRect.height; ++y) newRoi.convex[y].x -= minX1;
						}
//JFR_DEBUG("newRect end " << newRect);
//JFR_DEBUG("newRoi.rect end " << newRoi.rect);
						
						newRoi.checkIntegrity();
						return newRoi;
					}
				}
				
				/// intersection
				ConvexRoi& operator*=(const ConvexRoi &roi)
				{
					*this = *this * roi;
					return *this;
				}

				/// shift
				ConvexRoi  operator+ (const cv::Point &p) const
				{ 
					ConvexRoi res(*this);
					res.rect.x += p.x;
					res.rect.y += p.y;
					return res;
				}
				/// shift
				ConvexRoi  operator- (const cv::Point &p) const
				{ 
					ConvexRoi res(*this);
					res.rect.x -= p.x;
					res.rect.y -= p.y;
					return res;
				}
				
				/// shift
				ConvexRoi& operator+=(const cv::Point &p)
				{
					rect.x += p.x;
					rect.y += p.y;
					return *this;
				}
				/// shift
				ConvexRoi& operator-=(const cv::Point &p)
				{
					rect.x -= p.x;
					rect.y -= p.y;
					return *this;
				}
				
				/// scale
				void scaleTo(cv::Rect &newRect)
				{
// JFR_DEBUG("# scaleTo: oldRect " << rect << ", newRect " << newRect);
					cv::Point desiredOr(newRect.x, newRect.y);
					if (hasConvex)
					{
						// init new convex
						ConvexRoi newRoi(newRect);
						newRoi.initConvex();
						ConvexRoiImpl &newConvex = newRoi.convex;
// JFR_DEBUG("convex.size " << newConvex.size());
						double fw = newRect.width / (double)rect.width;
						newRect.x=std::floor(rect.x*fw);
// JFR_DEBUG("width fw " << fw);
						// first scale x
						for(ConvexRoiImpl::iterator it = convex.begin(); it != convex.end(); ++it)
						{
// JFR_DEBUG_BEGIN(); JFR_DEBUG_SEND("oldSlice " << *it);
/*							int x1 = std::floor((rect.x + it->x)*fw);
							int x2 = std::ceil((rect.x + it->x + it->w)*fw - 1);
							it->x = x1 - newRect.x;
							it->w = x2 - x1 + 1;
*/
							it->x = std::floor((rect.x + it->x)*fw) - newRect.x;
							it->w = std::ceil(it->w*fw);
							if (it->x+it->w > newRect.width) it->w = newRect.width-it->x;
// JFR_DEBUG_SEND(", newSlice x " << *it); JFR_DEBUG_END();
						}
						// then scale y
						int i = 0;
						newRoi.count_ = 0;
						double fh = newRect.height / (double)rect.height;
						newRect.y=std::floor(rect.y*fh);
						for(ConvexRoiImpl::iterator it = newConvex.begin(); it != newConvex.end(); ++it, ++i)
						{
							// same FIXME than in ellipse constructor
							//double jj = (i < newRect.height/2 ? (i+1)/fh : i/fh);
							int jj1 = std::floor(i/fh), jj2 = std::ceil((i+1)/fh);
							if (jj2 >= rect.height) jj2 = rect.height-1;
							int x1 = convex[jj1].x, x2 = convex[jj1].x+convex[jj1].w-1;
							for (int j = jj1+1; j <= jj2; ++j)
							{
								if (convex[j].x < x1) x1 = convex[j].x;
								int x2j = convex[j].x+convex[j].w-1;
								if (x2j > x2) x2 = x2j;
							}
							it->x = x1;
							it->w = x2-x1+1;
							newRoi.count_ += it->w;
// JFR_DEBUG("newSlice y " << *it << " by taking extrema between " << jj1 << " and " << jj2);
						}
#ifndef JFR_NDEBUG
						newRoi.rect = newRect;
						newRoi.checkIntegrity();
#endif
						*this = newRoi;
					} else
					{
						count_ = newRect.width*newRect.height;
					}
					rect = newRect;
					if (desiredOr.x >= 0) rect.x = desiredOr.x;
					if (desiredOr.y >= 0) rect.y = desiredOr.y;
					checkIntegrity();
				}
				
				
				/// scale
				ConvexRoi  operator* (double f) const
				{
					ConvexRoi res(*this);
					res *= f;
					return res; 
				}
				
				/// scale
				ConvexRoi& operator*=(double f)
				{
					cv::Rect newRect(-1,-1, std::ceil(rect.width *f), std::ceil(rect.height*f));
					scaleTo(newRect);
					return *this;
				}
				
				/// scale preserving center
				ConvexRoi& scale(double f)
				{
					cv::Point center(rect.x+rect.width/2, rect.y+rect.height/2);
					cv::Rect newRect(-1,-1, std::ceil(rect.width *f), std::ceil(rect.height*f));
					scaleTo(newRect);
					cv::Point newcenter(rect.x+rect.width/2, rect.y+rect.height/2);
					cv::Point shift = center-newcenter;
					*this += shift;
					return *this;
				}
				
				
			void draw()
			{
				std::cout << "rect: " << rect << std::endl;
				for(int i = 0; i < rect.height; ++i)
				{
					int x1 = convex[i].x;
					int x2 = convex[i].x+convex[i].w-1;
					for(int j = 0; j < x1; ++j) std::cout << ".";
					for(int j = x1; j <= x2; ++j) std::cout << "#";
					for(int j = x2+1; j < rect.width; ++j) std::cout << ".";
					std::cout << std::endl;
				}
			}
			
			void checkIntegrity()
			{
#ifndef JFR_NDEBUG
				JFR_ASSERT(rect.x > -1000000, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.x < 1000000, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.y > -1000000, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.y < 1000000, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.width >= 0, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.width < 1000000, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.height >= 0, "ConvexRoi::checkIntegrity failed");
				JFR_ASSERT(rect.height < 1000000, "ConvexRoi::checkIntegrity failed");
				
				if (hasConvex)
				{
					JFR_ASSERT((int)convex.size() == rect.height, "ConvexRoi::checkIntegrity failed");
					int minX1 = 2000000000, maxX2 = -2000000000;
					int recount = 0;
					for(ConvexRoiImpl::iterator it = convex.begin(); it != convex.end(); ++it)
					{
						JFR_ASSERT(std::abs(it->x) < 2000000000 && std::abs(it->w) < 2000000000, "ConvexRoi::checkIntegrity failed")
						if (it->x < minX1) minX1 = it->x;
						int x2 = it->x+it->w-1;
						if (x2 > maxX2) maxX2 = x2;
						recount += it->w;
					}
					if (rect.height > 0 && rect.width > 0)
					{
						JFR_ASSERT(minX1 == 0,"ConvexRoi::checkIntegrity failed");
						JFR_ASSERT(maxX2 == rect.width-1, "ConvexRoi::checkIntegrity failed");
					}
					JFR_ASSERT(recount == count_, "ConvexRoi::checkIntegrity failed");
				}
#endif
			}

			friend std::ostream& operator<<(std::ostream& os, ConvexRoi const& roi);
		};
		
		std::ostream& operator<<(std::ostream& os, ConvexRoi const& roi);
		
#if 0
		
		struct ConvexRange { int y, x, w; };

		class GeneralRoi: public std::list<ConvexRange>
		{
			private:
				
				typedef std::vector<iterator> RoiLines; // size +1
			
			private:
				
				RoiLines lines;
				
			public:
				GeneralRoi() {}
			public:
				// constructors
				Roi(Roi const & roi)
				{
					rect = roi.rect;
					if (roi.convex) { convex = new ConvexRoi(); *convex = *(roi.convex); } else convex = NULL;
					if (roi.mask) { mask = new image::Image(); *mask = *(roi.mask); } else mask = NULL;
				}
				Roi& operator=(Roi const & roi)
				{
					rect = roi.rect;
					delete convex; if (roi.convex) { convex = new ConvexRoi(); *convex = *(roi.convex); } else convex = NULL;
					delete mask; if (roi.mask) { mask = new image::Image(); *mask = *(roi.mask); } else mask = NULL;
				}
				Roi(cv::Rect &rect): rect(rect), convex(NULL), mask(NULL) {}
				Roi(int x, int y, int width, int height): rect(x,y,width,height), convex(NULL), mask(NULL) {}
				Roi(jblas::vec2 &x, jblas::sym_mat22 &P, double factor)
				{
					// TODO
				}
				Roi(int x, int y, image::Image &mask_): rect(x,y,mask_.width(),mask_.height()), convex(new ConvexRoi()), mask(new image::Image(mask_)
				{
					convex->resize(rect.height);
					for(int i = 0; i < rect.height; i++)
					{
						int j;
//						for(j = 0; j < rect.width; j++) if (mask(
					// TODO
					}
				}
				~Roi() { delete convex; delete mask; }
				
				iterator begin(int i) { return lines[i]; }
				iterator end(int i) { return lines[i+1]; }
				
				// accessors
				int top   () const { return rect.y; }
				int bottom() const { return rect.y+rect.height-1; }
				int left () const { return rect.x; }
				int right() const { return rect.x+rect.width-1; }
				int left (int i) const { return convex ? rect[i].first  : left() ; }
				int right(int i) const { return convex ? rect[i].second : right(); }
				bool test(int x, int y) const 
				{
					if (mask) return mask.at<uchar>(y-top(),x-left());
					if (convex) return y>=top() && y <= bottom() && x>=left(y-top()) && x<=right(y-top()); 
					       else return y>=top() && y <= bottom() && x>=left()        && x<=right(); 
				
				}
				double getRatio() { return 1; }
				
				// introspection
				bool isRect() const { return convex == NULL && mask == NULL; }
				bool isConvex() const { return mask == NULL; }
				
				// operations
				void scale(double factor) 
				{ 
					rect.x=(int)(rect.x*factor);
					rect.y=(int)(rect.y*factor);
					rect.width =(int)(rect.width *factor+0.999);
					rect.height=(int)(rect.height*factor+0.999);
				}
				void shift(int x, int y) 
				{ 
					rect.x += x; 
					rect.y += y; 
				}
				void add(Roi &roi) ///< A U B
				{
					// TODO
				}
				void remove(Roi &roi) ///< A \ B
				{
					// TODO
				}
				void intersect(Roi &roi) ///< A /\ B
				{
					// TODO
				}
		};
#endif


#if 0
		/**
		 * Region of Interest class.
		 *
		 * This class stores an upper-left corner and a size of a region of interest (ROI).
		 * \ingroup rtslam
		 */
		class RoiAbstract {
			
			public:
				virtual int top() const = 0;
				virtual int bottom() const = 0;
				virtual int left() const = 0; ///< min_i(left(i))
				virtual int right() const = 0; ///< max_i(right(i))
				virtual int left(int i) const = 0;
				virtual int right(int i) const = 0;
				virtual int width() const { return right()-left()+1; }
				virtual int height() const { return bottom()-top()+1; }
				/**
				Test whether some point is inside the convex ROI
				*/
				virtual bool test(int x, int y) const = 0;
				/**
				Test whether a point inside the convex ROI (given by top/bottom/left/right)
				is really inside the ROI
				*/
				virtual bool inTest(int x, int y) const = 0;
				virtual bool isConvex() const = 0;
				
				virtual RoiAbstract* clone() = 0;
				virtual void scale(double factor) = 0;
				virtual void shift(int x, int y) = 0;
//				virtual void remove(RoiAbstract &roi) = 0;

				std::ostream& operator<<(std::ostream& os)
				{
					os << "(" << left() << "," << top() << ";" << width() << "," << height() << ")";
					return os;
				}
		};
		

		class RoiRect
#if !USE_TEMPLATE_INLINE
			: public RoiAbstract
#endif
		{
			private:
				cv::Rect roi;
			public:
				RoiRect(cv::Rect &roi): roi(roi) {}
				RoiRect(int x, int y, int width, int height): roi(x,y,width,height) {}
				int top() const { return roi.y; }
				int bottom() const { return roi.y+roi.height-1; }
				int left() const { return roi.x; }
				int right() const { return roi.x+roi.width-1; }
				int left(int i) const { return left(); }
				int right(int i) const { return right(); }
				bool test(int x, int y) const { return y>=top() && y <= bottom() && x>=left() && x<=right(); }
				bool inTest(int x, int y) const { return true; }
				bool isConvex() const { return true; }
				RoiAbstract* clone() { return new RoiRect(roi); }
				void scale(double factor) { roi.x=(int)(roi.x*factor); roi.y=(int)(roi.y*factor); roi.width=(int)(roi.width*factor+0.999);  roi.height=(int)(roi.height*factor+0.999); }
				void shift(int x, int y) { roi.x += x; roi.y += y; }
		};
		
		class RoiEllipse
#if !USE_TEMPLATE_INLINE
			: public RoiAbstract
#endif
		{
			private:
				jblas::vec2 x;
				jblas::sym_mat22 P;
			public:
				RoiEllipse(jblas::vec2 &x, jblas::sym_mat22 &P) { /* TODO */ }
				int top() const { return 0;/* TODO */ }
				int bottom() const { return 0;/* TODO */ }
				int left() const { return 0;/* TODO */ }
				int right() const { return 0;/* TODO */ }
				int left(int i) const { return 0;/* TODO */ }
				int right(int i) const { return 0;/* TODO */ }
				bool test(int x, int y) const { return 0;/* TODO */ }
				bool inTest(int x, int y) const { return true; }
				bool isConvex() const { return true; }
		};
		
		class RoiConvex
#if !USE_TEMPLATE_INLINE
			: public RoiAbstract
#endif
		{
			private:
				int top_;
				int bottom_;
				std::vector<std::pair<int,int> > roi;
			public:
				RoiConvex(int maxHeight) { /* TODO */ }
				RoiConvex(jblas::vec2 &x, jblas::sym_mat22 &P) { /* TODO */ }
				RoiConvex(int top, int bottom, std::vector<int>) { /* TODO */ }
				void intFromEllipse(jblas::vec2 &x, jblas::sym_mat22 &P) { /* TODO */ }
				int top() const { return top_; }
				int bottom() const { return bottom_; }
				int left() const { return 0;/* TODO */ }
				int right() const { return 0;/* TODO */ }
				int left(int i) const { return roi[i].first; }
				int right(int i) const { return roi[i].second; }
				bool test(int x, int y) const { return y>=top_ && y<=bottom_ && x>=roi[y].first && x<=roi[y].second; }
				bool inTest(int x, int y) const { return true; }
				bool isConvex() const { return true; }
		};
		
		
		class RoiMask
#if !USE_TEMPLATE_INLINE
			: public RoiAbstract
#endif
		{
			private:
				cv::Rect roi;
				image::Image mask;
			public:
				RoiMask(int top, int left, image::Image &mask) { /* TODO */ }
				RoiMask(cv::Rect &roi, image::Image &mask) { /* TODO */ }
				int top() const { return roi.y; }
				int bottom() const { return roi.y+roi.height-1; }
				int left() const { return 0;/* TODO */ }
				int right() const { return 0;/* TODO */ }
				int left(int i) const { return roi.x; }
				int right(int i) const { return roi.x+roi.width-1; }
				bool test(int x, int y) const { return y>=top() && y <= bottom() && x>=left(y) && y<=right(y); }
				bool inTest(int x, int y) const { return mask.data(y)[x]; }
				bool isConvex() const { return false; }
		};

#endif 

//////////////////////////////////////////////////////////////////////
		
		
		class ROI;
		typedef boost::shared_ptr<ROI> roi_ptr_t;
		
		class ROI {
			public:
				int x, y;
				int width, height;
				ROI(const int ulx = 0, const int uly = 0, const int sx = 0, const int sy = 0) {
					x = ulx;
					y = uly;
					width = sx;
					height = sy;
				}
				veci2 downright() {
					veci2 dr;
					dr(0) = x + width - 1;
					dr(1) = y + height - 1;
					return dr;
				}
				veci2 upleft() {
					veci2 ul;
					ul(0) = x;
					ul(1) = y;
					return ul;
				}
				veci2 size() {
					veci2 sz;
					sz(0) = width;
					sz(1) = height;
					return sz;
				}
				void upleft(const veci & ul){
					x = ul(0);
					y = ul(1);
				}
				void downright(const veci & dr){
					width = dr(0) - x + 1;
					height = dr(1) - y + 1;
				}
				void size(const veci & sz){
					width = sz(0);
					height = sz(1);
				}

				friend std::ostream& operator <<(std::ostream & s, ROI & roi) {
					s << "ROI: " << std::endl;
					s << " .upper left  = " << roi.upleft() << std::endl;
					s << " .lower right = " << roi.downright() << std::endl;
					s << " .size        = " << roi.size() << std::endl;
					return s;
				}

		};
		

	}
}


#endif /* ROI_HPP_ */
